import os
import time


def poll_and_wait_until_true(func, max_seconds=10, sleep_between_trials=1, is_slurm=False):

    if is_slurm:
        sleep_between_trials = 5
        max_seconds = 100

    for i in range(0, max_seconds):
        if not func():
            time.sleep(sleep_between_trials)
        else:
            return False

    return True


def wait(the_task, is_slurm=False):
    max_seconds = 10 if not is_slurm else 100
    sleep_between_trials = 1 if not is_slurm else 3

    poll_and_wait_until_true(lambda: the_task.get_state().is_completed(), max_seconds, sleep_between_trials)

    if the_task.has_failed():
        raise Exception(f"{the_task} has failed")


def wait_for_step(the_task, step_number, is_slurm=False):
    max_seconds = 10 if not is_slurm else 100
    sleep_between_trials = 1 if not is_slurm else 3

    def check():
        if the_task.has_failed():
            raise Exception(f"{the_task} has failed")
        if the_task.next_step_number() == step_number:
            print("")
            return True

        return False

    exhausted_time = poll_and_wait_until_true(check, max_seconds, sleep_between_trials)

    if exhausted_time:
        raise Exception(f"exhausted time")


def load_file_as_string(f):
    with open(f) as f:
        return f.read()

def rewrite_file(file, content):
    with open(file, "w") as f:
        f.write(content)

def slurm_conf(dsl):
    return dsl.slurm(
        account="def-xroucou_cpu",
        sbatch_options=[
            "--time=0:5:00"
        ]
    )


def count(pipeline, pred):
    return sum([
        1 if pred(task) else 0
        for task in pipeline.tasks
    ])




import inspect
import os
import pathlib
import shutil
import yaml
import logging.config
from dry_pipe import DryPipe, script_lib


def test_suite_base_dir():
    return os.path.dirname(__file__)

def file_in_test_suite(f):
    return os.path.join(test_suite_base_dir(), f)

def setup_log_conf(log_file):
    log_conf_file = os.path.join(
        os.path.dirname(__file__),
        "logging-configs",
        log_file
    )

    with open(log_conf_file) as f:
        print(f"log config: {log_conf_file}")
        config = yaml.load(f, Loader=yaml.FullLoader)
        logging.config.dictConfig(config)


def log_4_debug_daemon_mode():
    setup_log_conf("debug-daemon-mode.yml")


class TestSandboxDir:

    def __init__(self, parent, other_func=None):

        s = inspect.stack()

        all_sandbox_dirs = os.path.join(
            os.path.dirname(__file__),
            "sandboxes"
        )

        if other_func is None:
            f = s[1].function
        else:
            f = other_func

        sandbox_dir = os.path.join(
            all_sandbox_dirs,
            f"{parent.__class__.__name__}.{f}"
        )

        self.sandbox_dir = sandbox_dir

        if other_func is None:
            self.delete_and_recreate_sandbox()


    def delete_and_recreate_sandbox(self):
        if os.path.exists(self.sandbox_dir):
            shutil.rmtree(self.sandbox_dir)

        pathlib.Path(self.sandbox_dir).mkdir(parents=True, exist_ok=False)

    def create_subdir(self, d):
        res = os.path.join(self.sandbox_dir, d)
        pathlib.Path(res).mkdir(parents=False, exist_ok=False)
        return res

    def pipeline_instance(self, template_pipeline):
        return template_pipeline.create_pipeline_instance(
            pipeline_instance_dir=self.sandbox_dir
        )

    def pipeline_instance_from_generator(
            self, task_generator, completed=False, env_vars=None, fail_silently=False, task_conf=None, remote_task_confs=None):
        from base_tests import test_containers_dir
        pi = DryPipe.create_pipeline(task_generator, remote_task_confs=remote_task_confs).create_pipeline_instance(
            pipeline_instance_dir=self.sandbox_dir,
            env_vars=env_vars,
            containers_dir=test_containers_dir(),
            task_conf=task_conf
        )

        copy_pre_existing_file_deps_from_code_dir(pi)

        if completed:
            pi.run_sync(fail_silently=fail_silently)

        return pi

    def init_pid_for_tests(self):
        def mkdir(d):
            d = os.path.join(self.sandbox_dir, d)
            pathlib.Path(d).mkdir(exist_ok=True)
            return d

        drypipe_dir = mkdir('.drypipe')
        mkdir('output')

        shutil.copy(script_lib.__file__, drypipe_dir)



def copy_pre_existing_file_deps_from_code_dir(pipeline_instance):

    instance_dir_for_test = pipeline_instance.pipeline_instance_dir

    src_pipeline = pipeline_instance.pipeline.create_pipeline_instance()

    file_deps = src_pipeline.pre_existing_file_deps()

    src_dir = src_pipeline.pipeline_instance_dir

    for f in file_deps:
        f = os.path.join(src_dir, f)
        if not os.path.exists(f):
            raise Exception(f"file not found {f}")
        shutil.copy(f, instance_dir_for_test)


def _copy_pre_existing_file_deps(src_pipeline, dst_pipeline_instance_dir):

    instance_dir_for_test = dst_pipeline_instance_dir

    src_pipeline = src_pipeline.pipeline.create_pipeline_instance()

    file_deps = src_pipeline.pre_existing_file_deps()

    src_dir = src_pipeline.pipeline_instance_dir

    for f in file_deps:
        f = os.path.join(src_dir, f)
        if not os.path.exists(f):
            raise Exception(f"file not found {f}")
        shutil.copy(f, instance_dir_for_test)


class A:
    pass


def zaz():
    TestSandboxDir(A())


if __name__ == '__main__':
    zaz()


def ensure_remote_dirs_dont_exist(pipeline_instance, clear_file_cache=False):

    remote_dir_for_test = pipeline_instance.instance_dir_base_name()

    for task_conf in pipeline_instance.remote_sites_task_confs():
        complete_remote_dir_for_test = f"{task_conf.remote_base_dir}/{remote_dir_for_test}"
        with task_conf.create_executer() as rex:
            rex.invoke_remote(f"rm -Rf {complete_remote_dir_for_test}")

            if clear_file_cache:
                rex.invoke_remote(f"rm -Rf {task_conf.remote_base_dir}/.file-cache")



import textwrap

from dry_pipe import DryPipe
from test_00_ground_level_tests.pipeline_defs import pipeline_exerciser_func1


def decorator_factory(argument):
    return lambda a: lambda: [argument, a]


@decorator_factory(argument=123)
def zaz(a, b, c, d):
    return f"{a} {b} {c} {d}"


@DryPipe.python_call(tests=[234, 432])
def zozo(a, b, c, test=None):
    return 432

if __name__ == '__main__':
    arg, func = zaz()
    print(func(4,3,2,1))
    print(f"arg={arg}")

    q = zozo()

    print("")




import glob
import importlib
import inspect
import json
import logging
import logging.config
import pathlib
import re
import shutil
import signal
import sys
import os
import time
import traceback

import click
from dry_pipe import DryPipe
from dry_pipe.internals import PythonCall
from dry_pipe.janitors import Janitor
from dry_pipe.monitoring import fetch_task_groups_stats
from dry_pipe.pipeline import PipelineInstance, Pipeline
from dry_pipe.pipeline_state import PipelineState
from dry_pipe.script_lib import env_from_sourcing, parse_in_out_meta, create_task_logger, iterate_out_vars_from, \
    write_out_vars
from dry_pipe.task_state import NON_TERMINAL_STATES

logger = logging.getLogger(__name__)


@click.group()
@click.option('-v', '--verbose')
@click.version_option()
@click.pass_context
def cli_group(ctx, verbose):
    ctx.ensure_object(dict)
    ctx.obj['verbose'] = verbose
    ctx.obj["ctx"] = ctx


def _bad_entrypoint(ctx):

    if "entrypoint" in ctx.obj and ctx.obj["entrypoint"] == "run_task":
        click.echo("Can't launch a pipeline with cli.run_task(), use cli.run(pipeline=<your pipeline>)", err=True)
        return True

    return False


def _configure_logging(log_conf_file):

    if log_conf_file is None:
        log_conf_file = os.environ.get("DRYPIPE_LOGGING_CONF")

    loggin_conf_in_current_dir = os.path.join(os.getcwd(), "drypipe-logging-conf.json")

    if log_conf_file is None and os.path.exists(loggin_conf_in_current_dir):
        log_conf_file = loggin_conf_in_current_dir

    if log_conf_file is not None:
        with open(log_conf_file) as f:
            conf = json.loads(f.read())

            handlers = conf["handlers"]

            for k, handler in handlers.items():
                handler_class = handler["class"]
                if handler_class == "logging.FileHandler":
                    filename = handler.get("filename")
                    if filename is None or filename == "":
                        raise Exception(f"logging.FileHandler '{k}' has no filename attribute in {log_conf_file}")
                    if "$" in filename:
                        filename = os.path.expandvars(filename)
                        handler["filename"] = os.path.expandvars(filename)

            logging.config.dictConfig(conf)
        logger.info("using logging conf: %s", log_conf_file)


@click.command()
@click.pass_context
@click.option('--instance-dir', type=click.Path(), default=None)
def clean(ctx, instance_dir):
    if _bad_entrypoint(ctx):
        return
    pipeline_func = ctx.obj["pipeline_func"]
    pipeline = _pipeline_from_pipeline_func(pipeline_func, instance_dir)
    # TODO: prompt
    pipeline.clean()


def _pipeline_from_pipeline_func(pipeline_func, instance_dir):

    pipeline_func_sig = inspect.signature(pipeline_func)

    nargs = len(pipeline_func_sig.parameters)

    if nargs == 0:

        if instance_dir is not None:
            raise Exception(f"when specifying instance_dir, task generator must take a dry_pipe.DryPipe.dsl() argument")

        return DryPipe.create_pipeline(pipeline_func)
    elif nargs == 1:
        #dsl = DryPipe.dsl_for(pipeline_func)
        pipeline_code_dir = DryPipe.pipeline_code_dir_for(pipeline_func)
        pipeline_instance_dir = pipeline_code_dir

        #if instance_dir is not None:
        #    dsl = dsl.with_defaults(pipeline_instance_dir=pipeline_code_dir)
        return DryPipe.create_pipeline(pipeline_func, pipeline_code_dir)
    else:
        m = inspect.getmodule(pipeline_func)
        raise Exception(f"function {pipeline_func.__name__} in {m.__file__} takes {nargs} arguments,\n"
                        f"it should take a instance created with dry_pipe.DryPipe.dsl()")


@click.command()
@click.option('-p', '--pipeline', help="a_module:a_func, a function that returns a pipeline.")
@click.option('--instance-dir', type=click.Path(), default=None)
@click.option('--regen-all', is_flag=True, default=False)
@click.option('--env', type=click.STRING, default=None)
def prepare(pipeline, instance_dir, regen_all, env):

    pipeline_instance = _pipeline_instance_creater(instance_dir, pipeline, env)()

    for task in pipeline_instance.tasks:

        if regen_all:
            task.prepare()
            print(f"regenerated {task.key}")


@click.command()
@click.option('-p', '--pipeline', help="a_module:a_func, a function that returns a pipeline.")
@click.option('--instance-dir', type=click.Path(), default=None)
@click.option('--dry-run', is_flag=False)
def change_keys(pipeline, instance_dir, dry_run):

    pipeline_instance = _pipeline_instance_creater(instance_dir, pipeline)()

    if dry_run:
        print("dry run, nothing will change !")

    for do_it, old_key, new_key in pipeline_instance.change_keys():

        if not dry_run:
            do_it()

        print(f"{old_key} -> {new_key}")

def _pipeline_instance_creater(instance_dir, module_func_pipeline, env=None):

    env_vars = None
    if env is not None:
        env_vars = {}
        for k_v in env.split(","):
            k, v = k_v.split("=")
            env_vars[k] = v

    if module_func_pipeline is None:
        module_func_pipeline = PipelineInstance.guess_pipeline_from_hints(instance_dir)

    pipeline = Pipeline.load_from_module_func(module_func_pipeline)

    PipelineInstance.write_hint_file_if_not_exists(instance_dir, module_func_pipeline)

    return lambda: pipeline.create_pipeline_instance(instance_dir, env_vars=env_vars)


@click.command()
@click.option('-p', '--pipeline', help="a_module:a_func, a function that returns a pipeline.")
@click.option('--instance-dir', type=click.Path(), default=None)
@click.option(
    '--group-by', type=click.STRING, default="by_task_type",
    help="""
    name of grouper in Drypipe.create_pipeline(...,task_groupers={'my_custom_grouper': group_func})
    """
)
def mon(pipeline, instance_dir, group_by):

    _configure_logging(None)

    from dry_pipe.cli_screen import CliScreen

    if instance_dir is None:
        instance_dir = os.getcwd()

    if pipeline is None:
        pipeline = PipelineInstance.guess_pipeline_from_hints(instance_dir)
        if pipeline is None:
            click.echo("must specify pipeline, ex: -p <aModule>:<aPipelineFunction>")

    #pipeline_instance = _pipeline_instance_creater(instance_dir, pipeline)()

    def _quit_listener(cli_screen):
        logger.debug("quit listener")
        logging.shutdown()
        cli_screen.cleanup_tty()
        os._exit(0)

    cli_screen = CliScreen(
        instance_dir,
        None,
        is_monitor_mode=True,
        quit_listener=_quit_listener,
        group_by=group_by
    )

    cli_screen.start()
    def _sigint(p1, p2):
        logger.info("received SIGINT")
        cli_screen.request_quit()

    signal.signal(signal.SIGINT, _sigint)
    signal.pause()

    logging.shutdown()
    exit(0)


@click.command()
@click.option('-p', '--pipeline', help="a_module:a_func, a function that returns a pipeline.")
@click.option('--instance-dir', type=click.Path(), default=None)
@click.option('--web-mon', is_flag=True)
@click.option('--port', default=5000, help="port for --web-mon")
@click.option('--bind', default="0.0.0.0", help="bind address for --web-mon")
@click.option('--clean', is_flag=True)
@click.option('--single', type=click.STRING, default=None,
              help="launches a single task, specified by the task key, in foreground, ex.: launch --single=TASK_KEY")
@click.option('--restart', is_flag=True, help="restart failed or killed tasks")
@click.option('--restart-failed', is_flag=True, help="restart failed tasks")
@click.option('--restart-killed', is_flag=True, help="restart killed tasks")
@click.option('--reset-failed', is_flag=True)
@click.option('--no-confirm', is_flag=True)
@click.option('--logging-conf', type=click.Path(), default=None, help="log configuration file (json)")
@click.option('--env', type=click.STRING, default=None)
@click.option(
    '--group-by', type=click.STRING, default="by_task_type",
    help="""
    name of grouper in Drypipe.create_pipeline(...,task_groupers={'my_custom_grouper': group_func})
    """
)
@click.option(
    '--regex-grouper', type=click.STRING, default=None,
    help="""
    <regex>,<format>
    example: --regex-grouper=.*\_(\w)\.(\d),group-{0}-{1}
    """
)
def run(
    pipeline, instance_dir, web_mon, port, bind,
    clean, single, restart, restart_failed, restart_killed, reset_failed,
    no_confirm, logging_conf, env, regex_grouper, group_by
):

    _configure_logging(logging_conf)

    pipeline_mod_func = pipeline

    if instance_dir is None:
        instance_dir = os.getcwd()

    if not os.path.isabs(instance_dir):
        instance_dir = os.path.abspath(os.path.expanduser(os.path.expandvars(instance_dir)))

    if PipelineState.from_pipeline_instance_dir(instance_dir, none_if_not_exists=True) is None:
        if not no_confirm:
            if not click.confirm(f"no pipeline instance exists in {instance_dir}, do you want to create one ?"):
                click.echo("no pipeline instance created, exiting.")
                return

    pipeline_instance = _pipeline_instance_creater(instance_dir, pipeline_mod_func, env)()

    if clean:
        pipeline_instance.clean()

    pipeline_instance.init_work_dir()

    if single is not None:
        launch_single(pipeline_instance, single)
        return

    pipeline_instance.regen_tasks_if_stale(force=True)

    for task in pipeline_instance.tasks:
        task_state = task.get_state()
        if task_state is not None:

            if (task_state.is_failed() or task_state.is_killed()) and not task.has_unsatisfied_deps():
                task.prepare()

                if restart:
                    task.re_queue()
                elif restart_failed and task_state.is_failed():
                    task.re_queue()
                elif restart_killed and task_state.is_killed():
                    task.re_queue()

    janitor = Janitor(
        pipeline_instance=pipeline_instance,
        min_sleep=0,
        max_sleep=5
    )

    janitor.start()
    janitor.start_remote_janitors()

    if web_mon:
        from dry_pipe.websocket_server import WebsocketServer
        from dry_pipe.server import PipelineUIServer
        server = PipelineUIServer.init_single_pipeline_instance(pipeline_instance.get_state())
        server.start()

        WebsocketServer.start(bind, port, instances_dir_if_has_local_janitor=pipeline_instance.pipeline_instance_dir)
        # Stopping threads break CTRL+C, let the threads die with the process exit
        #server.stop()
        #janitor.do_shutdown()
    else:
        from dry_pipe.cli_screen import CliScreen

        def _quit_listener(cli_screen):
            logger.debug("quit listener")
            logging.shutdown()
            cli_screen.cleanup_tty()
            os._exit(0)

        cli_screen = CliScreen(
            pipeline_instance.pipeline_instance_dir,
            None if regex_grouper is None else _display_grouper_func_from_spec(regex_grouper),
            is_monitor_mode=False,
            quit_listener=_quit_listener,
            group_by=group_by
        )

        cli_screen.start()
        def _sigint(p1, p2):
            logger.info("received SIGINT")

            if janitor is not None:
                janitor.request_shutdown()
            cli_screen.request_quit()

        signal.signal(signal.SIGINT, _sigint)
        signal.pause()

        logging.shutdown()
        exit(0)



    #os._exit(0)


def _display_grouper_func_from_spec(display_grouper):
    reg_fmt = display_grouper.split(",")
    if len(reg_fmt) != 2:
        raise Exception(
            f"bad format for arg --display-grouper, expects <regex>,<format> got {display_grouper}"
        )
    reg, fmt = reg_fmt
    try:
        r = re.compile(reg)
    except Exception as ex:
        raise Exception(f"--display-grouper has bad regex: {reg} ")

    fmt_insert_count = fmt.count("{")

    if r.groups != fmt_insert_count:
        raise Exception(
            f"regex match group and placeholders in format must match, got: {r.groups} != {fmt_insert_count}"
        )

    def _regex_display_grouper(task_key):
        m = r.search(task_key)
        if m is None:
            return "<Other Tasks>"
        else:
            gr = m.groups()
            if len(gr) != fmt_insert_count:
                return "<Other Tasks>"
            else:
                return fmt.format(*m.groups())
    return  _regex_display_grouper


def launch_single(pipeline, single):

    task = pipeline.tasks.get(single)

    if task is None:
        raise Exception(f"pipeline has no task with key '{single}'")

    task.re_queue()
    task.prepare()
    task_state = task.get_state()
    task_state.transition_to_launched(task.task_conf.create_executer(), task, wait_for_completion=True)
    task_state = task.get_state()
    task_state.transition_to_completed(task)

    print(f"Task {single} completed.")


@click.command()
@click.pass_context
@click.option('--instance-dir', type=click.Path(), default=None)
@click.option('--single', type=click.STRING, default=None,
              help="apply to single task, specified by the task key, in foreground, ex.: launch --single=TASK_KEY")
def requeue(ctx, single, instance_dir):

    if _bad_entrypoint(ctx):
        return

    pipeline_func = ctx.obj["pipeline_func"]

    pipeline = _pipeline_from_pipeline_func(pipeline_func, instance_dir)

    if single is None:
        raise Exception(f"unimplemented.")

    requeue_single(pipeline, single)


def requeue_single(pipeline, single):

    task = pipeline.tasks.get(single)

    if task is None:
        raise Exception(f"pipeline has no task with key '{single}'")

    task.re_queue()
    task.prepare()

    print(f"Task {single} requeued")

def _parse_instances_dir_to_pipelines(instances_dir_to_pipelines, env=None):

    for d_to_p in instances_dir_to_pipelines.split(","):

        instances_dir, mod_func = d_to_p.split("=")

        if not os.path.exists(instances_dir):
            pathlib.Path(instances_dir).mkdir(parents=True)

        pipeline = Pipeline.load_from_module_func(mod_func)
        pipeline.env_vars = env
        yield instances_dir, pipeline, mod_func

def _version():

    frame = inspect.currentframe()
    f_back = frame.f_back if frame is not None else None
    f_globals = f_back.f_globals if f_back is not None else None
    # break reference cycle
    # https://docs.python.org/3/library/inspect.html#the-interpreter-stack
    del frame

    package_name = f_globals.get("__name__")

    if package_name == "__main__":
        package_name = f_globals.get("__package__")

    if package_name:
        package_name = package_name.partition(".")[0]

    try:
        from importlib import metadata
    except ImportError:
        # Python < 3.8
        import importlib_metadata as metadata
    try:
        return metadata.version(package_name)
    except metadata.PackageNotFoundError:
        raise RuntimeError("Could not get version")


@click.command()
@click.pass_context
@click.option(
    '--instances-dir-to-pipelines',
    type=click.STRING,
    help="see 'watch' command"
)
@click.option(
    '-o',
    type=click.STRING,
    help="output file"
)
def dump_remote_task_confs(ctx, instances_dir_to_pipelines, o):

    def g():
        for instances_dir, pipeline, mod_func in _parse_instances_dir_to_pipelines(instances_dir_to_pipelines):
            yield {
                "constructor_func": mod_func,
                "pipeline_instances_dir": instances_dir,
                "remote_task_confs": [
                    tc.as_json() for tc in pipeline.remote_task_confs
                ]
            }

    if False:
        res = json.dumps({
            "dry_pipe_version": _version(),
            "watched_pipelines_config": list(g())
        }, indent = 2)
    else:
        res = list(g())

    if o is None:
        print(res)
    else:
        with open(o, "w") as o_f:
            o_f.write(res)


@click.command()
@click.pass_context
@click.option(
    '--instances-dir-to-pipelines',
    type=click.STRING,
    required=True,
    help="""
    dir1=module1:func1,...,dirN=moduleN:funcN    
      + moduleX:funcX is a function that returns a dry_pipe.Pipeline    
      + dirX is the parent directory where all pipeline_instance_dirs of PipelineX will reside  
    """
)
@click.option('--env', type=click.STRING, default=None)
def watch(ctx, instances_dir_to_pipelines, env):

    _configure_logging(None)

    janitor = Janitor(
        pipeline_instances_iterator=Pipeline.pipeline_instances_iterator({
            k: v
            for k, v, _ in _parse_instances_dir_to_pipelines(instances_dir_to_pipelines, env)
        })
    )

    janitor.start()

    janitor.start_remote_janitors()

    def _sigterm(p1, p2):
        logger.info("received SIGTERM")
        janitor.request_shutdown()
        logging.shutdown()
        time.sleep(3)
        os._exit(0)

    signal.signal(signal.SIGTERM, _sigterm)
    signal.pause()






def _validate_task_generator_callable_with_single_dsl_arg_and_get_module_file(task_generator):

    a = inspect.signature(task_generator)

    f = os.path.dirname(os.path.abspath(inspect.getmodule(task_generator).__file__))

    if len(a.parameters) != 1:
        raise Exception(f"function given to cli.run() should take a singe DryPipe.dsl() as argument, and yield tasks" +
                        f" was given {len(a.parameters)} args")

    return f


@click.command()
@click.pass_context
@click.option(
    '--instances-dir-to-pipelines',
    type=click.STRING,
    required=True,
    help="""
    dir1=module1:func1,...,dirN=moduleN:funcN    
      + moduleX:funcX is a function that returns a dry_pipe.Pipeline    
      + dirX is the parent directory where all pipeline_instance_dirs of PipelineX will reside  
    """
)
@click.option('--bind', default="0.0.0.0", help="bind address for --web-mon")
@click.option('--port', default=5000)
def serve_ui(ctx, instances_dir_to_pipelines, bind, port):
    from dry_pipe.websocket_server import WebsocketServer

    _configure_logging(None)

    def g():
        for t in instances_dir_to_pipelines.split(","):

            instances_dir, mod_func = t.split("=")

            if not os.path.exists(instances_dir):
                pathlib.Path(instances_dir).mkdir(parents=True)

            pipeline = Pipeline.load_from_module_func(mod_func)
            yield instances_dir, pipeline

    WebsocketServer.start(bind, port, dict(g()))

    os._exit(0)


def _func_from_mod_func(mod_func):

    mod, func_name = mod_func.split(":")

    if not mod.startswith("."):
        module = importlib.import_module(mod)
    else:
        module = importlib.import_module(mod[1:])

    python_task = getattr(module, func_name, None)
    if python_task is None:
        raise Exception(f"function {func_name} not found in module {mod}")

    return python_task


@click.command()
@click.pass_context
@click.argument('test-case-mod-func', type=click.STRING)
def test(ctx, test_case_mod_func):

    idx = None

    if test_case_mod_func.count(":") == 1:
        mod, func_name = test_case_mod_func.split(":")
    elif test_case_mod_func.count(":") == 2:
        mod, func_name, idx = test_case_mod_func.split(":")
        idx = int(idx)
    else:
        raise Exception(
            f"expected an argument of the form <module>:<function_name>, or <module>:<function_name>:<number>"
        )

    module = importlib.import_module(mod)

    python_call = getattr(module, func_name, None)

    if python_call is None:
        raise Exception(f"function {func_name} not found in module {mod}")

    if not isinstance(python_call, PythonCall):
        raise Exception(f"function {func_name} should be decorated with @DryPipe.python_call(tests=[...])")

    if len(python_call.tests) == 0:
        click.echo(f"python_task {test_case_mod_func} has no tests")
        return

    def run_test(python_test_case):
        init_func = python_test_case.get("init_func")

        if init_func is not None:
            init_func()

        res = python_call.func(*python_test_case["args"], test=python_test_case)

        expects = python_test_case.get("expects")
        if expects is not None:
            if res != expects:
                raise Exception(f"expected {json.dumps(expects, indent=2)}\ngot:\n{json.dumps(res, indent=2)}")

    c = 0

    if idx is None:
        for python_test_case in python_call.tests:
            run_test(python_test_case)
            print(f"test passed: {mod}:{func_name}:{c}")
            c += 1
    else:
        run_test(python_call.tests[idx])
        print(f"test passed: {test_case_mod_func}")


@click.command()
@click.pass_context
@click.argument('mod-func', type=click.STRING)
@click.option('--task-env', type=click.Path(exists=True, dir_okay=False))
def call(ctx, mod_func, task_env):

    python_task = _func_from_mod_func(mod_func)

    env = os.environ

    if task_env is not None:
        # or if pipeline provided --task=key
        env = env_from_sourcing(task_env)

    control_dir = env["__control_dir"]

    task_logger = create_task_logger(control_dir)

    var_type_dict = {}

    meta = {
        k : v
        for k, v in env.items()
        if k.startswith("__meta_")
    }

    for producing_task_key, var_metas, _ in parse_in_out_meta(meta):
        #if producing_task_key != "":
        #    continue

        for var_meta in var_metas:
            name_in_producing_task, var_name, typez = var_meta
            var_type_dict[var_name] = typez


    def get_and_parse_arg(k, allow_none):
        v = env.get(k)
        if allow_none and v is None:
            return None
        if v is None and k != "test":
            tk = os.environ['__task_key']
            raise Exception(
                f"Task {tk} called {mod_func} with None assigned to arg {k}\n" +
                f"make sure task has var {k} declared in it's consumes clause. Ex:\n" +
                f"  dsl.task(key={tk}).consumes({k}=...)"
            )
        typez = var_type_dict.get(k)
        if typez == "int":
            return int(v)
        elif typez == "str":
            return v
        elif typez == "float":
            return float(v)
        elif typez == "glob_expression":

            class GlobExpression:
                def __call__(self, *args, **kwargs):
                    return glob.glob(os.path.expandvars(v))

                def __str__(self):
                    return os.path.expandvars(v)

            return GlobExpression()

        return v

    args_tuples = [
        (k, get_and_parse_arg(k, allow_none=False))
        for k, v in python_task.signature.parameters.items()
        if not k == "kwargs"
    ]

    args = [v for _, v in args_tuples]

    if "kwargs" not in python_task.signature.parameters:
        kwargs = {}
    else:

        args_names = [k for k, _ in args_tuples]

        task_logger.debug("args list: %s", args_names)

        kwargs = {
            k : get_and_parse_arg(k, allow_none=True)
            for k, _ in var_type_dict.items()
            if k not in args_names
        }

        for k, v in meta.items():
            if v.startswith("file") and k not in kwargs:
                f = k[7:]
                if f not in args_names:
                    kwargs[f] = env.get(f)


    task_logger.info("will invoke PythonCall: %s(%s,%s)", mod_func, args, kwargs)

    out_vars = None

    try:
        out_vars = python_task.func(* args, ** kwargs)
    except Exception as ex:
        traceback.print_exc()
        logging.shutdown()
        exit(1)

    if out_vars is not None and not isinstance(out_vars, dict):
        raise Exception(
            f"function {python_task.mod_func()} called by task {os.environ.get('__task_key')} {type(out_vars)}" +
            f"@DryPipe.python_call() can only return a python dict, or None"
        )

    try:
        if out_vars is not None:
            prev_out_vars = dict(iterate_out_vars_from(os.environ["__output_var_file"]))

            for k, v in out_vars.items():
                try:
                    v = json.dumps(v)
                    prev_out_vars[k] = v
                except TypeError as ex0:
                    print(
                        f"task call {control_dir}.{mod_func} returned var {k} of unsupported type: {type(v)}" +
                        " only primitive types are supported.",
                        file=sys.stderr
                    )
                    logging.shutdown()
                    exit(1)

            write_out_vars(prev_out_vars)

        #for pid_file in glob.glob(os.path.join(control_dir, "*.pid")):
        #    os.remove(pid_file)

    except Exception as ex:
        traceback.print_exc()
        logging.shutdown()
        exit(1)

    logging.shutdown()


@click.command()
@click.pass_context
@click.option('--downstream-of', type=click.STRING)
@click.option('--not-matching', type=click.STRING)
@click.option('--matching', type=click.STRING)
@click.option('--instance-dir', type=click.Path(), default=None)
def reset(ctx, downstream_of, not_matching, matching, instance_dir):

    pipeline_func = ctx.obj["pipeline_func"]

    pipeline = _pipeline_from_pipeline_func(pipeline_func, instance_dir)

    if matching is not None or not_matching is not None:

        if matching is not None and not_matching is not None:
            raise Exception("--matching and --not-matching can't be both specified")

        def do_reset(key):
            if matching is not None:
                return re.match(matching, key)
            return not re.match(not_matching, key)

        for task in pipeline.tasks:
            if do_reset(task.key):
                if click.confirm(f"reset {task.key} ?"):
                    task.clean()
                    print(f"{task.key} reset")

    elif downstream_of is not None:

        after_task_including = pipeline.tasks.get(downstream_of)

        if after_task_including is None:
            raise Exception(f"pipeline has no task with key '{downstream_of}'")

        def all_upstream_deps(task):
            yield task
            for upstream_task, _1, _2 in task.upstream_deps_iterator():
                for t in all_upstream_deps(upstream_task):
                    yield t

        raise Exception("not implemented")
    else:
        if click.confirm("reset all tasks ?"):
            print("boom")

@click.command()
@click.pass_context
@click.option('--instance-dir', type=click.Path(), default=None)
@click.option('--no-header', is_flag=True, default=False)
@click.option('--units',  type=click.Choice(["seconds", "minutes", "hours"]), default="minutes")
def stats(ctx, instance_dir, no_header, units):

    if instance_dir is not None:
        the_stats = fetch_task_groups_stats(instance_dir, no_header, units)
    else:
        pipeline_func = ctx.obj["pipeline_func"]
        pipeline = _pipeline_from_pipeline_func(pipeline_func, instance_dir)
        the_stats = fetch_task_groups_stats(pipeline.pipeline_instance_dir, no_header, units)

    for stat_row in the_stats:
        print("\t".join([str(v) for v in stat_row]))


@click.command()
@click.pass_context
@click.option('-p', '--pipeline', help="a_module:a_func, a function that returns a pipeline.")
def prepare_remote_sites(ctx, pipeline):

    pipeline_spec = pipeline

    if pipeline_spec is None:
        click.echo("must specify pipeline, ex: -p <aModule>:<aPipelineFunction>")
        return

    pipeline = Pipeline.load_from_module_func(pipeline_spec)

    from rich.console import Console
    from rich.panel import Panel
    console = Console()

    for f, msg in pipeline.prepare_remote_sites_funcs():
        console.print(Panel(f"[turquoise2]{msg}"))
        with console.status("rsync upload in progress..."):
            f()

    console.print(f"done.")

@click.command()
@click.pass_context
def status(ctx):

    raise Exception(f"not implemented")

    if _bad_entrypoint(ctx):
        return

    pipeline_func = ctx.obj["pipeline_func"]

    pipeline = DryPipe.create_pipeline(pipeline_func)
    console = Console()
    console.print(_status_table(pipeline))


def _register_commands():
    cli_group.add_command(run)
    cli_group.add_command(mon)
    cli_group.add_command(prepare)
    cli_group.add_command(call)
    cli_group.add_command(test)
    cli_group.add_command(status)
    cli_group.add_command(watch)
    cli_group.add_command(serve_ui)
    cli_group.add_command(clean)
    cli_group.add_command(stats)
    cli_group.add_command(reset)
    cli_group.add_command(requeue)
    cli_group.add_command(dump_remote_task_confs)
    cli_group.add_command(prepare_remote_sites)

def run_cli():
    _register_commands()
    cli_group()

if __name__ == '__main__':
    run_cli()



import glob
import os
import unittest

import pipeline_with_code_and_container_upsync
import pipelines_with_external_file_deps_and_remote_task
from dry_pipe import TaskConf, DryPipe
from pipeline_with_two_remote_sites import create_pipeline_generator_two_remote_sites
from pipelines_with_remote_tasks import dag_gen_fileset_output
from test_04_remote_ssh_tasks import pipeline_with_remote_tasks
from test_04_remote_ssh_tasks.pipeline_with_remote_tasks import \
    complete_and_validate_pipeline_instance
from test_utils import TestSandboxDir, ensure_remote_dirs_dont_exist


class RemoteTaskTests1(unittest.TestCase):

    def _test_prepare_code_dir_remote_site(self, d, with_container):

        gen, validator, tc = pipeline_with_code_and_container_upsync.dag_gen(
            self,
            with_container
        )

        pipeline_code_dir = os.path.dirname(os.path.dirname(__file__))

        pipeline_instance = DryPipe.create_pipeline(
            gen,
            pipeline_code_dir=pipeline_code_dir,
            remote_task_confs=[tc]
        ).create_pipeline_instance(
            pipeline_instance_dir=d.sandbox_dir,
            task_conf=tc
        )

        ensure_remote_dirs_dont_exist(pipeline_instance)

        pipeline_instance.pipeline.prepare_remote_sites()

        pipeline_instance.run_sync()

        validator(pipeline_instance)

    def test_prepare_code_dir_remote_site(self):
        d = TestSandboxDir(self)
        self._test_prepare_code_dir_remote_site(d, None)

    def _test_prepare_code_dir_remote_site_with_container(self):
        d = TestSandboxDir(self)
        pipeline_code_dir = os.path.dirname(os.path.dirname(__file__))
        container = os.path.join(pipeline_code_dir, "tests", "containers", "singularity-test-container.sif")
        self._test_prepare_code_dir_remote_site(d, container)

    def test_remote_file_cache(self):

        d = TestSandboxDir(self)

        gen, validator = pipelines_with_external_file_deps_and_remote_task.dag_gen(TaskConf(
            executer_type="process",
            ssh_specs=f"maxl@ip32.ccs.usherbrooke.ca:~/.ssh/id_rsa",
            remote_base_dir="/nfs3_ib/ip32-ib/home/maxl/drypipe-tests"
        ), self)


        pipeline_instance = d.pipeline_instance_from_generator(gen)

        ensure_remote_dirs_dont_exist(pipeline_instance, clear_file_cache=True)

        pipeline_instance.run_sync()

        validator(pipeline_instance)

    def test_remote_task_with_fileset_output(self):
        d = TestSandboxDir(self)

        gen, validator = dag_gen_fileset_output(TaskConf(
            executer_type="process",
            ssh_specs=f"maxl@ip32.ccs.usherbrooke.ca:~/.ssh/id_rsa",
            remote_base_dir="/nfs3_ib/ip32-ib/home/maxl/drypipe-tests",
            remote_containers_dir="/nfs3_ib/ip32-ib/home/maxl/drypipe-tests/containers"
        ), self)

        pipeline_instance = d.pipeline_instance_from_generator(gen)

        ensure_remote_dirs_dont_exist(pipeline_instance)

        pipeline_instance.run_sync()

        validator(pipeline_instance)

    def test_remote_tasks_basics(self):

        d = TestSandboxDir(self)

        pipeline_instance = d.pipeline_instance(pipeline_with_remote_tasks.create_pipeline_with_remote_tasks(
            TaskConf(
                executer_type="process",
                ssh_specs=f"maxl@ip32.ccs.usherbrooke.ca:~/.ssh/id_rsa",
                remote_base_dir="/nfs3_ib/ip32-ib/home/maxl/drypipe-tests",
                remote_containers_dir="/nfs3_ib/ip32-ib/home/maxl/drypipe-tests/containers"
            )
        ))

        ensure_remote_dirs_dont_exist(pipeline_instance)

        complete_and_validate_pipeline_instance(pipeline_instance, self)


    def test_remote_tasks_using_remote_code_dir(self):

        d = TestSandboxDir(self)

        pipeline_instance = d.pipeline_instance(pipeline_with_remote_tasks.create_pipeline_with_remote_tasks(
            TaskConf(
                executer_type="process",
                ssh_specs=f"maxl@ip32.ccs.usherbrooke.ca:~/.ssh/id_rsa",
                remote_base_dir="/nfs3_ib/ip32-ib/home/maxl/drypipe-tests",
                remote_pipeline_code_dir="/nfs3_ib/ip32-ib/home/maxl/drypipe-tests/remote_code_dir_test_04"
            )
        ))

        ensure_remote_dirs_dont_exist(pipeline_instance)

        pipeline_instance.pipeline.prepare_remote_sites()

        complete_and_validate_pipeline_instance(pipeline_instance, self)

    def test_pipeline_with_two_remote_sites(self):

        d = TestSandboxDir(self)
        tc1 = TaskConf(
            executer_type="process",
            ssh_specs=f"maxl@ip32.ccs.usherbrooke.ca:~/.ssh/id_rsa",
            remote_base_dir="/nfs3_ib/ip32-ib/home/maxl/drypipe-tests",
            remote_containers_dir="dummy1"
        )

        tc2 = TaskConf(
            executer_type="process",
            ssh_specs="maxl@ip29.ccs.usherbrooke.ca:~/.ssh/id_rsa",
            remote_base_dir="/home/maxl/drypipe_tests",
            remote_containers_dir="dummy2"
        )

        pipeline_instance = d.pipeline_instance_from_generator(create_pipeline_generator_two_remote_sites(tc1, tc2))

        for task_conf in pipeline_instance.remote_sites_task_confs():
            if task_conf.ssh_specs.startswith("maxl@ip32"):
                task_conf_ip32 = task_conf
            elif task_conf.ssh_specs.startswith("maxl@ip29"):
                task_conf_ip29 = task_conf
            else:
                raise Exception("!")

        ensure_remote_dirs_dont_exist(pipeline_instance)

        with task_conf_ip32.create_executer() as remote_executor_ip32:
            with task_conf_ip29.create_executer() as remote_executor_ip29:

                pipeline_instance.init_work_dir()

                remote_executor_ip29.prepare_remote_instance_directory(pipeline_instance, task_conf_ip29)
                remote_executor_ip32.prepare_remote_instance_directory(pipeline_instance, task_conf_ip32)

                pipeline_instance.run_sync()

                with open(os.path.join(d.sandbox_dir, "output", "t3", "f3.txt")) as f:
                    self.assertEqual(
                        {1, 2},
                        {int(s) for s in f.read().split() if s.strip() != ""}
                    )

                def fetch_containers_dir_in_pipeline_env(remote_executor, task_conf):
                    f = os.path.join(
                        task_conf.remote_base_dir,
                        "RemoteTaskTests1.test_pipeline_with_two_remote_sites",
                        ".drypipe",
                        "pipeline-env.sh"
                    )

                    try:
                        content = remote_executor.fetch_remote_file_content(f)
                        res = [
                            l.split("=")[1]
                            for l in content.split("\n")
                            if l.startswith("export __containers_dir")
                        ]
                        self.assertEqual(len(res), 1)
                        return res[0].strip()
                    except FileNotFoundError as fnfe:
                        raise Exception(f"{fnfe} while fetching {f} from {task_conf.ssh_specs}")

                self.assertEqual(fetch_containers_dir_in_pipeline_env(remote_executor_ip32, tc1), "dummy1")
                self.assertEqual(fetch_containers_dir_in_pipeline_env(remote_executor_ip29, tc2), "dummy2")

class RemoteTaskTests2(unittest.TestCase):

    def test_remote_tasks_basics_with_container(self):

        d = TestSandboxDir(self)

        pipeline_instance = d.pipeline_instance(pipeline_with_remote_tasks.create_pipeline_with_remote_tasks(
            TaskConf(
                executer_type="process",
                ssh_specs="maxl@ip29.ccs.usherbrooke.ca:~/.ssh/id_rsa",
                remote_base_dir="/home/maxl/drypipe_tests",
                container="singularity-test-container.sif",
                remote_containers_dir="/home/maxl/containers",
                command_before_task="module add apptainer",
            )
        ))

        ensure_remote_dirs_dont_exist(pipeline_instance)

        complete_and_validate_pipeline_instance(pipeline_instance, self)


class RemoteTaskTestsWithSlurm(unittest.TestCase):

    def test_remote_tasks_with_slurm(self):

        d = TestSandboxDir(self)

        cedar = TaskConf(
            executer_type="slurm",
            slurm_account="def-xroucou_cpu",
            sbatch_options=[
                "--time=0:5:00"
            ],
            ssh_specs=f"maxl@cedar.computecanada.ca:~/.ssh/id_ed25519",
            remote_base_dir="/home/maxl/drypipe-tests"
        )

        ip32 = TaskConf(
            executer_type="slurm",
            slurm_account="def-xroucou_cpu",
            sbatch_options=[
                "--time=0:5:00"
            ],
            ssh_specs=f"maxl@ip32.ccs.usherbrooke.ca:~/.ssh/id_rsa",
            remote_base_dir="/nfs3_ib/ip32-ib/home/maxl/drypipe-tests"
        )

        pipeline_instance = d.pipeline_instance(pipeline_with_remote_tasks.create_pipeline_with_remote_tasks(
            ip32
        ))

        ensure_remote_dirs_dont_exist(pipeline_instance)

        complete_and_validate_pipeline_instance(pipeline_instance, self)

    def test_remote_tasks_with_slurm_and_container(self):

        d = TestSandboxDir(self)

        pipeline_instance = d.pipeline_instance(pipeline_with_remote_tasks.create_pipeline_with_remote_tasks(
            TaskConf(
                executer_type="slurm",
                slurm_account="def-xroucou_cpu",
                sbatch_options=[
                    "--time=0:5:00"
                ],
                ssh_specs="maxl@ip29.ccs.usherbrooke.ca:~/.ssh/id_rsa",
                container="singularity-test-container.sif",
                remote_containers_dir="/home/maxl/containers",
                remote_base_dir="/home/maxl/drypipe_tests"
            )
        ))

        ensure_remote_dirs_dont_exist(pipeline_instance)

        complete_and_validate_pipeline_instance(pipeline_instance, self)



import glob
import logging
import os
import pathlib

from dry_pipe.script_lib import PortablePopen

TASK_ACTIONS = {

    "restart": ["failed", "timed-out", "completed"],
    "pause": [],
    "kill": []
}

PIPELINE_ACTIONS = {
    "kill-all-and-pause": [],
    "pause-all": [],
    "reset-all": []
}


ALL_TASK_ACTIONS = [a for a, l in TASK_ACTIONS.items()]


logger = logging.getLogger(__name__)

class TaskAction:

    @staticmethod
    def fetch_all_actions(pipeline_instance_dir):
        for action_file in glob.glob(os.path.join(pipeline_instance_dir, ".drypipe", "*", "action.*")):
            yield TaskAction.load_from_file(action_file)

    @staticmethod
    def load_from_task_control_dir(d):

        action_files = list(glob.glob(os.path.join(d, "action.*")))
        c = len(action_files)

        if c > 1:
            raise Exception(f"{c} action files found in {d}")
        elif c == 0:
            return None

        return TaskAction.load_from_file(action_files[0])

    @staticmethod
    def load_from_file(action_file):

        action, action_name = os.path.basename(action_file).split(".")

        if action_name.startswith("restart"):
            step = 0
            if "-" in action_name:
                action_name, step = action_name.split("-")

            return TaskAction(action_name, action_file, int(step))

        return TaskAction(action_name, action_file)

    @staticmethod
    def load_from_task_state(task_state):
        return TaskAction.load_from_task_control_dir(task_state.control_dir())

    @staticmethod
    def submit(pipeline_instance_dir, task_key, action_name, step=None):

        TaskAction.validate_action_name(action_name)

        pipeline_work_dir = os.path.join(pipeline_instance_dir, ".drypipe")

        control_dir = os.path.join(pipeline_work_dir, task_key)

        existing_action = TaskAction.load_from_task_control_dir(control_dir)

        if existing_action is not None:
            raise Exception(
                f"can't submit action to {control_dir}, action '{existing_action.action_name}' already exists."
            )

        if action_name == "restart":
            if step is not None:
                action_name = f"{action_name}-{step}"

        p = os.path.join(control_dir, f"action.{action_name}")

        pathlib.Path(p).touch(exist_ok=False)

        pathlib.Path(os.path.join(pipeline_work_dir, "sync")).touch(exist_ok=True)

    @staticmethod
    def validate_action_name(action_name):
        if action_name not in ALL_TASK_ACTIONS:
            raise Exception(f"invalid action_name: {action_name}")

    def __init__(self, action_name, file, step=None):

        TaskAction.validate_action_name(action_name)

        self.action_name = action_name
        self.file = file
        task_control_dir = os.path.dirname(file)
        self.task_key = os.path.basename(task_control_dir)
        self.step = step
        self.pipeline_instance_dir = os.path.dirname(os.path.dirname(task_control_dir))

    def delete(self):
        os.remove(self.file)

    def is_restart(self):
        return self.action_name == "restart"

    def is_pause(self):
        return self.action_name == "pause"

    def is_kill(self):
        return self.action_name == "kill"

    def is_deactivated(self):
        return self.is_pause()

    def unpause(self):
        self.delete()

    def do_it(self, task, executor):

        if self.action_name == "pause":
            return

        task_state = task.get_state()
        #executor = executer_getter.get_executer(task.task_conf)

        if self.is_restart():
            self._do_restart(task, task_state, executor)
        elif self.is_kill():
            self._do_kill(task, task_state, executor)

        self.delete()

    def _do_pause(self):
        pathlib.Path(os.path.join(
            os.path.dirname(self.file),
            "action.pause"
        )).touch()

    def _do_restart(self, task, task_state, executor):

        if task_state.is_failed() or task_state.is_completed() or \
                task_state.is_timed_out() or task_state.is_completed_unsigned() or task_state.is_killed():

            if task.is_remote():
                logger.info("will clear state of remote task %s", task.key)
                executor.clear_remote_task_state(task)


            logger.info("will requeue task %s", task.key)
            task_state.transition_to_queued(increment_retry=False, step=self.step, force=True)
            task.reset_logs()
        else:
            raise Exception(f"restart not yet implemented for tasks in state {task_state.state_name}")

    def _do_kill(self, task, task_state, executor):

        if task_state.is_failed() or task_state.is_completed() or \
                task_state.is_timed_out() or task_state.is_completed_unsigned():
            return

        if task.is_remote():
            if executor.slurm is not None:
                executor.kill_slurm_task(task)
                executor.clear_remote_task_state(task)
                task_state.transition_to_killed()
            else:
                task.log_control_error(
                    f"Warning: remote killing of non slurm tasks not yet implemented, use scancel or kill -9")
        else:

            from dry_pipe.internals import Local
            from dry_pipe.internals import Slurm

            if isinstance(executor, Local):
                executor.kill_task(task.key)
                task_state.transition_to_killed()
            elif isinstance(executor, Slurm):

                slurm_job_id_file = os.path.join(
                    task_state.v_abs_control_dir(),
                    "slurm_job_id"
                )

                if os.path.exists(slurm_job_id_file):
                    with open(slurm_job_id_file) as f:
                        slurm_job_id = f.read().strip()

                        call_scancel(slurm_job_id)
                else:
                    task.log_control_error(
                        f"Warning: can't find slurm job_id, file {slurm_job_id_file} missing.")

        self._do_pause()


def call_scancel(job_id):

    with PortablePopen(
        ["scancel", job_id]
    ) as p:
        p.wait_and_raise_if_non_zero()



import unittest
from datetime import datetime

import test_helpers as th
from dry_pipe.monitoring import fetch_task_groups_stats
from dry_pipe.task_state import TaskState
from test_regressions_pipelines.regression_pipeline_defs import pipeline_consuming_upstream_vars_with_same_name, \
    correct_timestamps_in_task_history
from test_utils import TestSandboxDir


class RegressionTests(unittest.TestCase):

    def test_pipeline_consuming_upstream_vars_with_same_name(self):

        d = TestSandboxDir(self)

        pipeline = d.pipeline_instance_from_generator(
            pipeline_consuming_upstream_vars_with_same_name, completed=True
        )

        self.assertEqual(
            th.count(pipeline, lambda task: task.has_completed()),
            3
        )


    def test_correct_timestamps_in_task_history(self):

        d = TestSandboxDir(self)

        pipeline_for_correct_timestamps_in_task_history = \
            d.pipeline_instance_from_generator(correct_timestamps_in_task_history, completed=True)

        sleep1, sleep2_3 = pipeline_for_correct_timestamps_in_task_history.tasks

        v1 = dict(sleep1.iterate_out_vars())
        v2_3 = dict(sleep2_3.iterate_out_vars())

        def p(dt):
            return datetime.strptime(dt[1:-1], '%Y-%m-%dT%H:%M:%S.%f')

        start1 = p(v1["start_t_1"])
        end1 = p(v1["end_t_1"])

        start2 = p(v2_3["start_t_2"])
        end2 = p(v2_3["end_t_2"])

        start3 = p(v2_3["start_t_3"])
        end3 = p(v2_3["end_t_3"])

        def filter_step_start_step_complete(task):
            for row in task.get_state().load_history_rows():
                state_name = row[0]
                timestamp = row[1]
                if state_name in ["step-started", "step-completed"]:
                    yield TaskState.parse_history_timestamp(timestamp)

        h_start1, h_end1 = list(filter_step_start_step_complete(sleep1))

        h_start2, h_end2, h_start3, h_end3 = list(filter_step_start_step_complete(sleep2_3))

        def f(d):
            return datetime.strftime(d, '%S.%f')

        self.assertTrue(
            h_start1 != h_end1,
            f"expected {f(h_end1)} to be at least 1 second later than {f(h_start1)}"
        )

        self.assertTrue(
            start1 != h_end1,
            f"expected {f(h_end1)} to be at least 1 second later than {f(h_start1)}"
        )

        self.assertTrue(
            end1 <= h_end1,
            f"expected {f(end1)} to be earlier or equal to {f(h_end1)}"
        )

        self.assertTrue(
            h_end1 < h_start2,
            f"expected {f(h_end1)} to be at least 1 second later than {f(h_start1)}"
        )

        def ensure1sec(d1, d2):
            self.assertTrue(
                (d2 - d1).seconds >= 1,
                f"expected on second between {f(d1)} and {f(d2)}"
            )

        def sanity_check(h_start, start, end, h_end):
            ensure1sec(h_start, end)
            ensure1sec(start, h_end)

        sanity_check(h_start1, start1, end1, h_end1)
        sanity_check(h_start2, start2, end2, h_end2)
        sanity_check(h_start3, start3, end3, h_end3)

        expected_timestamp_ordering = [
            h_start1, start1, end1, h_end1,
            h_start2, start2, end2, h_end2,
            h_start3, start3, end3, h_end3
        ]

        self.assertEqual(
            expected_timestamp_ordering,
            list(sorted(expected_timestamp_ordering))
        )

    def test_time_stats(self):

        d = TestSandboxDir(self)

        pipeline_for_correct_timestamps_in_task_history = \
            d.pipeline_instance_from_generator(correct_timestamps_in_task_history, completed=True)


        stats = list(
            fetch_task_groups_stats(
                pipeline_for_correct_timestamps_in_task_history.pipeline_instance_dir, no_header=True
            )
        )

        self.assertEqual(len(stats), 3)

        for step_name, min_t, max_t, total_t, avg_t, s_dev in stats:
            self.assertEqual(min_t, max_t)
            self.assertTrue(avg_t >= 1)



import os


def __env_var(name, default=None):

    name = f"DRYPIPE_{name}"

    if name in os.environ:
        return os.environ[name]

    return default


def log_config():
    return __env_var("LOG_CONFIG")


def test_remote_user():
    return __env_var("TEST_REMOTE_USER", "maxl")


def test_remote_host():
    return __env_var("TEST_REMOTE_HOST", "ip32.ccs.usherbrooke.ca")


def test_remote_ssh_key_path():
    return __env_var("TEST_REMOTE_SSH_SECRET_KEY_PATH", "/home/maxou/.ssh/id_rsa")


def test_remote_dir():
    #return __env_var("TEST_REMOTE_DIR", "/home/maxl/drypipe-tests")
    return __env_var("TEST_REMOTE_DIR", "/nfs3_ib/ip32-ib/home/maxl/drypipe-tests")


def test_remote_connection_string():
    return f"{test_remote_user()}@{test_remote_host()}:{test_remote_dir()}:{test_remote_ssh_key_path()}"


def test_slurm_group():
    return __env_var("TEST_SLURM_GROUP", "def-xroucou_cpu")